#nullable disable

using System.Text.Json.Serialization;
using System.Text.Json;

public record CurrencyExchangeRatesResult {
	public CurrencyExchangeRates Data { get; set; }
}

public record CurrencyExchangeRates {
	[JsonPropertyName("currency")]
	public string CurrencyId { get; set; }
	public IReadOnlyDictionary<string, decimal> Rates { get; set; }
}

public class Exchange {
	public string Id;
	public string Url;
	public Func<string, decimal> ParsePrice;
}

public class Program {

	private static List<Exchange> s_listReal = new List<Exchange> {
		new Exchange { Id = "USD", Url = "https://api.coinbase.com/v2/exchange-rates?currency=USD", ParsePrice = ParseRateReal },
		new Exchange { Id = "EUR", Url = "https://api.coinbase.com/v2/exchange-rates?currency=EUR", ParsePrice = ParseRateReal },
		new Exchange { Id = "GBP", Url = "https://api.coinbase.com/v2/exchange-rates?currency=GBP", ParsePrice = ParseRateReal },
	};
	private static SemaphoreSlim s_listRealSemaphore = new SemaphoreSlim(initialCount: 1, maxCount: 1);

	private static JsonSerializerOptions s_webJsonOptions = new(JsonSerializerDefaults.Web);

	private static decimal ParseRateReal(string json) {
		var result = JsonSerializer.Deserialize<CurrencyExchangeRatesResult>(json, s_webJsonOptions);
		return result.Data.Rates["CZK"];
	}

	private static HttpClient s_httpClient = new HttpClient();

	private static async Task<decimal> GetAvgRateAsync(List<Exchange> list) {
		s_listRealSemaphore.Wait();

		decimal sum = 0;
		foreach (var ex in list) {
			string json = await s_httpClient.GetStringAsync(ex.Url);
			sum += ex.ParsePrice(json);
		}
		var avg = sum / list.Count;

		s_listRealSemaphore.Release();
		return avg;
	}

	public record AvgRateResult(decimal AvgRate);

	public record RatePost(string Id, string Url);

	public static void Main(string[] args) {
		#region Initialize Kestrel
		var builder = WebApplication.CreateBuilder(args);

		// Add services to the container.
		// Learn more about configuring Swagger/OpenAPI at https://aka.ms/aspnetcore/swashbuckle
		builder.Services.AddEndpointsApiExplorer();
		builder.Services.AddSwaggerGen();

		var app = builder.Build();

		// Configure the HTTP request pipeline.
		if (app.Environment.IsDevelopment()) {
			app.UseSwagger();
			app.UseSwaggerUI();
		}

		app.UseHttpsRedirection();
		#endregion

		app.MapGet("/rates/average", () => {
			var avgRateFuture = GetAvgRateAsync(s_listReal);
			// Do some further processing
			return new AvgRateResult(avgRateFuture.Result);
		})
		.WithOpenApi();

		app.MapPost("/rates", (RatePost ratePost) => {
			s_listRealSemaphore.Wait();
			s_listReal.Add(new Exchange { Id = ratePost.Id, Url = ratePost.Url, ParsePrice = ParseRateReal });
			s_listRealSemaphore.Release();
		})
		.WithOpenApi();

		app.MapDelete("/rates/{id}", (string id) => {
			lock (s_listReal) {
				s_listReal.RemoveAll(x => x.Id == id);
			}
		})
		.WithOpenApi();

		app.Run();
	}
}