﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Drawing;

using System.Net.Http;
using System.Net.Http.Json;

using System.Text.Json.Serialization;
using System.Text.Json;

using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;

namespace HttpClientParallelRequestsBenchmark {

	public record CurrencyExchangeRatesResult {
		public CurrencyExchangeRates Data { get; set; }
	}

	public record CurrencyExchangeRates {
		[JsonPropertyName("currency")]
		public string CurrencyId { get; set; }
		public IReadOnlyDictionary<string, decimal> Rates { get; set; }
	}

	public class AvgRateBenchmarks {

		internal class Exchange {
			public string Id;
			public string Url;
			public Func<string, decimal> ParsePrice;
		}

		static HttpClient s_httpClient = new HttpClient();

		static SemaphoreSlim s_listSemaphore = new SemaphoreSlim(1, 1);

		internal static List<Exchange> s_listFake = new List<Exchange> {
			new Exchange { Id = "USD", Url = "https://api.coinbase.com/v2/exchange-rates?currency=USD", ParsePrice = ParseRateFake },
			new Exchange { Id = "EUR", Url = "https://api.coinbase.com/v2/exchange-rates?currency=EUR", ParsePrice = ParseRateFake },
			new Exchange { Id = "GBP", Url = "https://api.coinbase.com/v2/exchange-rates?currency=GBP", ParsePrice = ParseRateFake },
		};

		static decimal ParseRateFake(string json) {
			return 100;
		}

		internal static List<Exchange> s_listReal = new List<Exchange> {
			new Exchange { Id = "USD", Url = "https://api.coinbase.com/v2/exchange-rates?currency=USD", ParsePrice = ParseRateReal },
			new Exchange { Id = "EUR", Url = "https://api.coinbase.com/v2/exchange-rates?currency=EUR", ParsePrice = ParseRateReal },
			new Exchange { Id = "GBP", Url = "https://api.coinbase.com/v2/exchange-rates?currency=GBP", ParsePrice = ParseRateReal },
		};


		static JsonSerializerOptions s_webJsonOptions = new(JsonSerializerDefaults.Web);

		static decimal ParseRateReal(string json) {
			var result = JsonSerializer.Deserialize<CurrencyExchangeRatesResult>(json, s_webJsonOptions);
			return result.Data.Rates["CZK"];
		}

		internal static async Task<decimal> GetAvgRateDummyMonitorAsync(List<Exchange> list) {
			Monitor.Enter(list);
			Monitor.Exit(list);

			decimal sum = 0;
			foreach (var ex in list) {
				string json = await s_httpClient.GetStringAsync(ex.Url);
				sum += ex.ParsePrice(json);
			}
			var avg = sum / list.Count;
			
			// Monitor.Exit(list);
			return avg;
		}

		internal static async Task<decimal> GetAvgRateSemaphoreSlimAsync(List<Exchange> list) {
			await s_listSemaphore.WaitAsync();

			decimal sum = 0;
			foreach (var ex in list) {
				string json = await s_httpClient.GetStringAsync(ex.Url);
				sum += ex.ParsePrice(json);
			}
			var avg = sum / list.Count;

			s_listSemaphore.Release();
			return avg;
		}

		internal static async Task<decimal> GetAvgRateParallelAsync(List<Exchange> list) {
			Exchange[] copy;
			lock (list) {
				copy = list.ToArray();
			}

			Task<decimal>[] tasks = new Task<decimal>[copy.Length];

			for (int i = 0; i < copy.Length; i++) {
				var ex = copy[i];
				tasks[i] = s_httpClient.GetStringAsync(ex.Url).ContinueWith(t => ex.ParsePrice(t.Result));
			}

			var results = await Task.WhenAll(tasks);

			decimal sum = 0;
			foreach (var result in results) {
				sum += result;
			}

			var avg = sum / list.Count;
			return avg;
		}

		internal static async Task<decimal> GetAvgRateJustRequestsParallelAsync(List<Exchange> list) {
			Exchange[] copy;
			lock (list) {
				copy = list.ToArray();
			}

			Task<string>[] tasks = new Task<string>[copy.Length];

			for (int i = 0; i < copy.Length; i++) {
				var ex = copy[i];
				tasks[i] = s_httpClient.GetStringAsync(ex.Url);
			}

			var results = await Task.WhenAll(tasks);

			decimal sum = 0;
			for (int i = 0; i < results.Length; i++) {
				var result = copy[i].ParsePrice(results[i]);
				sum += result;
			}

			var avg = sum / list.Count;
			return avg;
		}

		[Benchmark(Baseline = true)]
		public decimal AwaitSeqDummyMonitorFake() {
			var task = GetAvgRateDummyMonitorAsync(s_listFake);
			return task.Result;
		}

		[Benchmark]
		public decimal AwaitSeqSemaphoreSlimFake() {
			var task = GetAvgRateSemaphoreSlimAsync(s_listFake);
			return task.Result;
		}

		[Benchmark]
		public decimal ParallelAwaitWhenAllFake() {
			var task = GetAvgRateParallelAsync(s_listFake);
			return task.Result;
		}

		[Benchmark]
		public decimal AwaitSeqDummyMonitorReal() {
			var task = GetAvgRateDummyMonitorAsync(s_listReal);
			return task.Result;
		}

		[Benchmark]
		public decimal AwaitSeqSemaphoreSlimReal() {
			var task = GetAvgRateSemaphoreSlimAsync(s_listReal);
			return task.Result;
		}

		[Benchmark]
		public decimal ParallelAwaitWhenAllReal() {
			var task = GetAvgRateParallelAsync(s_listReal);
			return task.Result;
		}

		[Benchmark]
		public decimal ParallelAwaitWhenAllJsonSeqReal() {
			var task = GetAvgRateJustRequestsParallelAsync(s_listReal);
			return task.Result;
		}
	}

	class Program {
		static void Main(string[] args) {
			BenchmarkRunner.Run<AvgRateBenchmarks>();
			//var task = AvgRateBenchmarks.GetAvgRateSemaphoreSlimAsync(AvgRateBenchmarks.s_listReal);
			//task.Wait();
			//Console.WriteLine($"Result: {task.Result}");
		}
	}
}
