#nullable disable

using System.Text.Json.Serialization;
using System.Text.Json;

public record CurrencyExchangeRatesResult {
	public CurrencyExchangeRates Data { get; set; }
}

public record CurrencyExchangeRates {
	[JsonPropertyName("currency")]
	public string CurrencyId { get; set; }
	public IReadOnlyDictionary<string, decimal> Rates { get; set; }
}

public class Exchange {
	public string Id;
	public string Url;
	public Func<string, decimal> ParsePrice;
}

public class Program {

	private static List<Exchange> s_listReal = new List<Exchange> {
		new Exchange { Id = "USD", Url = "https://api.coinbase.com/v2/exchange-rates?currency=USD", ParsePrice = ParseRateReal },
		new Exchange { Id = "EUR", Url = "https://api.coinbase.com/v2/exchange-rates?currency=EUR", ParsePrice = ParseRateReal },
		new Exchange { Id = "GBP", Url = "https://api.coinbase.com/v2/exchange-rates?currency=GBP", ParsePrice = ParseRateReal },
	};
	private static SemaphoreSlim s_listRealSemaphore = new SemaphoreSlim(initialCount: 1, maxCount: 1);

	private static JsonSerializerOptions s_webJsonOptions = new(JsonSerializerDefaults.Web);

	private static decimal ParseRateReal(string json) {
		var result = JsonSerializer.Deserialize<CurrencyExchangeRatesResult>(json, s_webJsonOptions);
		return result.Data.Rates["CZK"];
	}

	private static HttpClient s_httpClient = new HttpClient();

	private static async Task<decimal> GetAvgRateAsync(List<Exchange> list) {
		Exchange[] copy;
		try {
			await s_listRealSemaphore.WaitAsync();
			copy = list.ToArray();
		} finally {
			s_listRealSemaphore.Release();
		}

		Task<decimal>[] tasks = new Task<decimal>[copy.Length];

		for (int i = 0; i < copy.Length; i++) {
			var ex = copy[i];
			tasks[i] = s_httpClient.GetStringAsync(ex.Url).ContinueWith(t => ex.ParsePrice(t.Result));
		}

		var results = await Task.WhenAll(tasks);

		decimal sum = 0;
		foreach (var result in results) {
			sum += result;
		}

		var avg = sum / copy.Length;
		return avg;

	}

	public record AvgRateResult(decimal AvgRate);

	public record RatePost(string Id, string Url);

	public static void Main(string[] args) {
		#region Initialize Kestrel
		var builder = WebApplication.CreateBuilder(args);

		// Add services to the container.
		// Learn more about configuring Swagger/OpenAPI at https://aka.ms/aspnetcore/swashbuckle
		builder.Services.AddEndpointsApiExplorer();
		builder.Services.AddSwaggerGen();

		var app = builder.Build();

		// Configure the HTTP request pipeline.
		if (app.Environment.IsDevelopment()) {
			app.UseSwagger();
			app.UseSwaggerUI();
		}

		// Do not redirect us to HTTPS -> unencrypted HTTP connections should stay HTTP - allows us to use "bombardier" without problem without a valid HTTPS certificate:
		// app.UseHttpsRedirection();
		#endregion

		app.MapGet("/rates/average", async () => {
			var avgRateFuture = GetAvgRateAsync(s_listReal);
			// Do some further processing
			return new AvgRateResult(await avgRateFuture);
		})
		.WithOpenApi();

		app.MapPost("/rates", async (RatePost ratePost) => {
			try {
				await s_listRealSemaphore.WaitAsync();
				s_listReal.Add(new Exchange { Id = ratePost.Id, Url = ratePost.Url, ParsePrice = ParseRateReal });
			} finally {
				s_listRealSemaphore.Release();
			}
		})
		.WithOpenApi();

		app.MapDelete("/rates/{id}", async (string id) => {
			try {
				await s_listRealSemaphore.WaitAsync();
				s_listReal.RemoveAll(x => x.Id == id);
			} finally {
				s_listRealSemaphore.Release();
			}
		})
		.WithOpenApi();

		app.MapDelete("/rates-experimental-v2/{id}", async (string id) => {
			Console.WriteLine($"experimental-v2: Request #{GetRequestIndexSafe()}: Deleting {id} ...");
			try {
				await s_listRealSemaphore.WaitAsync();
				await Task.Delay(1);
				s_listReal.RemoveAll(x => x.Id == id);
			} finally {
				s_listRealSemaphore.Release();
			}
		})
		.WithOpenApi();

		app.Run();
	}

	public static long GetRequestIndexSafe() {
		var incrementedIndex = Interlocked.Increment(ref s_requestIndex);
		return incrementedIndex;
	}

	private static long s_requestIndex = 0;
}